local class			= require "middleclass"
local plugin		= require "bunkerweb.plugin"
local utils     	= require "bunkerweb.utils"
local cachestore	= require "bunkerweb.cachestore"
local cjson			= require "cjson"
local ipmatcher		= require "resty.ipmatcher"

local greylist = class("greylist", plugin)

function greylist:initialize()
	-- Call parent initialize
	plugin.initialize(self, "greylist")
	-- Check if redis is enabled
	local use_redis, err = utils.get_variable("USE_REDIS", false)
	if not use_redis then
		self.logger:log(ngx.ERR, err)
	end
	self.use_redis = use_redis == "yes"
	-- Decode lists
	if ngx.get_phase() ~= "init" and self.variables["USE_GREYLIST"] == "yes" then
		local lists, err = self.datastore:get("plugin_greylist_lists")
		if not lists then
			self.logger:log(ngx.ERR, err)
		else
			self.lists = cjson.decode(lists)
		end
		local kinds = {
			["IP"] = {},
			["RDNS"] = {},
			["ASN"] = {},
			["USER_AGENT"] = {},
			["URI"] = {}
		}
		for kind, _ in pairs(kinds) do
			for data in self.variables["GREYLIST_" .. kind]:gmatch("%S+") do
				table.insert(self.lists[kind], data)
			end
		end
	end
	-- Instantiate cachestore
	self.cachestore = cachestore:new(self.use_redis)
end

function greylist:init()
	-- Check if init is needed
	local init_needed, err = utils.has_variable("USE_GREYLIST", "yes")
	if init_needed == nil then
		return self:ret(false, "can't check USE_GREYLIST variable : " .. err)
	end
	if not init_needed or self.is_loading then
		return self:ret(true, "init not needed")
	end
	-- Read greylists
	local greylists = {
		["IP"] = {},
		["RDNS"] = {},
		["ASN"] = {},
		["USER_AGENT"] = {},
		["URI"] = {},
	}
	local i = 0
	for kind, _ in pairs(greylists) do
		local f, err = io.open("/var/cache/bunkerweb/greylist/" .. kind .. ".list", "r")
		if f then
			for line in f:lines() do
				table.insert(greylists[kind], line)
				i = i + 1
			end
			f:close()
		end
	end
	-- Load them into datastore
	local ok, err = self.datastore:set("plugin_greylist_lists", cjson.encode(greylists))
	if not ok then
		return self:ret(false, "can't store greylist list into datastore : " .. err)
	end
	return self:ret(true, "successfully loaded " .. tostring(i) .. " bad IP/network/rDNS/ASN/User-Agent/URI")
end

function greylist:access()
	-- Check if access is needed
	if self.variables["USE_GREYLIST"] ~= "yes" then
		return self:ret(true, "greylist not activated")
	end
	-- Check the caches
	local checks = {
		["IP"] = "ip" .. ngx.ctx.bw.remote_addr
	}
	if ngx.ctx.bw.http_user_agent then
		checks["UA"] = "ua" .. ngx.ctx.bw.http_user_agent
	end
	if ngx.ctx.bw.uri then
		checks["URI"] = "uri" .. ngx.ctx.bw.uri
	end
	local already_cached = {
		["IP"] = false,
		["URI"] = false,
		["UA"] = false
	}
	for k, v in pairs(checks) do
		local cached, err = self:is_in_cache(v)
		if not cached and err ~= "success" then
			self.logger:log(ngx.ERR, "error while checking cache : " .. err)
		elseif cached and cached ~= "ok" then
			return self:ret(true, k .. " is in cached greylist", utils.get_deny_status())
		end
		if cached then
			already_cached[k] = true
		end
	end
	-- Check lists
	if not self.lists then
		return self:ret(false, "lists is nil")
	end
	-- Perform checks
	for k, v in pairs(checks) do
		if not already_cached[k] then
			local greylisted, err = self:is_greylisted(k)
			if greylisted == nil then
				self.logger:log(ngx.ERR, "error while checking if " .. k .. " is greylisted : " .. err)
			else
				local ok, err = self:add_to_cache(self:kind_to_ele(k), greylisted or "ok")
				if not ok then
					self.logger:log(ngx.ERR, "error while adding element to cache : " .. err)
				end
				if greylisted == "ko" then
					return self:ret(true, k .. " is not in greylist", utils.get_deny_status())
				end
			end
		end
	end

	-- Return
	return self:ret(true, "greylisted")
end

function greylist:preread()
	return self:access()
end

function greylist:kind_to_ele(kind)
	if kind == "IP" then
		return "ip" .. ngx.ctx.bw.remote_addr
	elseif kind == "UA" then
		return "ua" .. ngx.ctx.bw.http_user_agent
	elseif kind == "URI" then
		return "uri" .. ngx.ctx.bw.uri
	end
end

function greylist:is_greylisted(kind)
	if kind == "IP" then
		return self:is_greylisted_ip()
	elseif kind == "URI" then
		return self:is_greylisted_uri()
	elseif kind == "UA" then
		return self:is_greylisted_ua()
	end
	return false, "unknown kind " .. kind
end

function greylist:is_greylisted_ip()
	-- Check if IP is in greylist
	local ipm, err = ipmatcher.new(self.lists["IP"])
	if not ipm then
		return nil, err
	end
	local match, err = ipm:match(ngx.ctx.bw.remote_addr)
	if err then
		return nil, err
	end
	if match then
		return true, "ip"
	end

	-- Check if rDNS is needed
	local check_rdns = true
	if self.variables["GREYLIST_RDNS_GLOBAL"] == "yes" and not ngx.ctx.bw.ip_is_global then
		check_rdns = false
	end
	if check_rdns then
		-- Get rDNS
		local rdns, err = utils.get_rdns(ngx.ctx.bw.remote_addr)
		-- Check if rDNS is in greylist
		if rdns then
			for i, suffix in ipairs(self.lists["RDNS"]) do
				if rdns:sub(-#suffix) == suffix then
					return true, "rDNS " .. suffix
				end
			end
		end
	end

	-- Check if ASN is in greylist
	if ngx.ctx.bw.ip_is_global then
		local asn, err = utils.get_asn(ngx.ctx.bw.remote_addr)
		if not asn then
			return nil, err
		end
		for i, bl_asn in ipairs(self.lists["ASN"]) do
			if bl_asn == tostring(asn) then
				return true, "ASN " .. bl_asn
			end
		end
	end

	-- Not greylisted
	return false, "ko"
end

function greylist:is_greylisted_uri()
	-- Check if URI is in greylist
	for i, uri in ipairs(self.lists["URI"]) do
		if ngx.ctx.bw.uri:match(uri) then
			return true, "URI " .. uri
		end
	end
	-- URI is not greylisted
	return false, "ko"
end

function greylist:is_greylisted_ua()
	-- Check if UA is in greylist
	for i, ua in ipairs(self.lists["USER_AGENT"]) do
		if ngx.ctx.bw.http_user_agent:match(ua) then
			return true, "UA " .. ua
		end
	end
	-- UA is not greylisted
	return false, "ko"
end

function greylist:is_in_cache(ele)
	local ok, data = self.cachestore:get("plugin_greylist_" .. ngx.ctx.bw.server_name .. ele)
	if not ok then
		return false, data
	end 
	return true, data
end

function greylist:add_to_cache(ele, value)
	local ok, err = self.cachestore:set("plugin_greylist_" .. ngx.ctx.bw.server_name .. ele, value, 86400)
	if not ok then
		return false, err
	end 
	return true
end

return greylist